"""
Test on unambiguous graph compression that prioritize nodes with
the greater degree.

"""
import unittest
import tempfile

from powergrasp import commons
from powergrasp import recipes
from powergrasp import solving
from powergrasp import config


class TestHighPriorityFirstUnambiguousCompression(unittest.TestCase):
    """This class yields many subtests about High Priority First (HPF) method
    and unambiguous compression results.

    The test case and associated results are defined in setUp() method.
    Bubble outputs are comparated, excluding comments and empty lines.

    The unambiguous compression is important : bubble comparison is sensitive
    to powernodes numbers and related.
    However, it resists to multiple connected components.

    """

    def setUp(self):
        self.fuzzy_test_cases = {
            'prio_deg.lp'             : RESULT_FUZZY_PRIODEG,
            'empty.lp'                : '',
        }
        self.test_cases = {
            'prio_deg.lp'             : RESULT_PRIODEG,
            'prio_deg_str.lp'         : RESULT_PRIODEG_STR,
            'empty.lp'                : '',
        }


    def unified_bubble(self, bubble_lines):
        """Return the set of comparable lines found in given bubble lines"""
        # filter out comments and blank lines
        return set(line.strip() for line in bubble_lines
                   if not line.startswith('#') and len(line.strip()) > 0)

    def unified_bubble_from_file(self, bubble_file):
        """Return the set of comparable lines found in given bubble file"""
        with open(bubble_file) as fd:
            return self.unified_bubble(fd)

    def assert_equality(self, input_filename, expected_bubble, fuzzy=False):
        """call assertEqual to compare given bubble and bubble generated by
        compressing given graph."""
        # create the temp file to be write by the compression
        tmp = tempfile.NamedTemporaryFile('w', delete=False)
        tmp.close()
        # launch the compression itself, then compare result with expected
        with self.subTest(filename=input_filename):
            cfg = config.Configuration.fields_for_high_priority_first(
                infile=input_filename,
                outfile=tmp.name,
                outformat='bbl',
                loglevel='CRITICAL',
                fuzzy=fuzzy,
            )
            recipes.powergraph(cfg=cfg)
            expected = self.unified_bubble(expected_bubble)
            found = self.unified_bubble_from_file(tmp.name)
            self.assertEqual(expected, found)


    def test_fuzzy_cases(self):
        for filename, expected in self.fuzzy_test_cases.items():
            self.assert_equality(
                commons.test_case(filename),
                expected.split('\n'),
                fuzzy=True,
            )

    def test_cases(self):
        for filename, expected in self.test_cases.items():
            self.assert_equality(
                commons.test_case(filename),
                expected.split('\n'),
            )


RESULT_PRIODEG = """
NODE\ta
NODE\th
NODE\tg
NODE\td
NODE\tf
NODE\tb
NODE\tc
NODE\te
IN\td\tPWRN-a-3-2
IN\te\tPWRN-a-3-2
IN\ta\tPWRN-a-1-1
IN\th\tPWRN-a-1-1
IN\tf\tPWRN-a-1-1
IN\tg\tPWRN-a-1-1
IN\tPWRN-a-3-2\tPWRN-a-2-2
IN\tb\tPWRN-a-2-1
IN\tc\tPWRN-a-2-1
EDGE\ta\tPWRN-a-3-2\t1.0
EDGE\tPWRN-a-1-1\ts\t1.0
EDGE\tPWRN-a-2-1\tPWRN-a-2-2\t1.0
"""

RESULT_PRIODEG_STR = """
NODE\tna
NODE\tnh
NODE\tng
NODE\tnd
NODE\tnf
NODE\tnb
NODE\tnc
NODE\tne
IN\tnd\tPWRN-na-3-2
IN\tne\tPWRN-na-3-2
IN\tna\tPWRN-na-1-1
IN\tnh\tPWRN-na-1-1
IN\tnf\tPWRN-na-1-1
IN\tng\tPWRN-na-1-1
IN\tPWRN-na-3-2\tPWRN-na-2-2
IN\tnb\tPWRN-na-2-1
IN\tnc\tPWRN-na-2-1
EDGE\tna\tPWRN-na-3-2\t1.0
EDGE\tPWRN-na-1-1\tns\t1.0
EDGE\tPWRN-na-2-1\tPWRN-na-2-2\t1.0
"""

RESULT_FUZZY_PRIODEG = """
NODE\th
NODE\ta
NODE\tb
NODE\tf
NODE\te
NODE\tg
NODE\td
NODE\tc
IN\tf\tPWRN-a-2-1
IN\tg\tPWRN-a-2-1
IN\th\tPWRN-a-2-1
IN\te\tPWRN-a-1-2
IN\td\tPWRN-a-1-2
IN\ta\tPWRN-a-1-1
IN\tc\tPWRN-a-1-1
IN\tb\tPWRN-a-1-1
EDGE\ta\ts\t1.0
EDGE\tPWRN-a-2-1\ts\t1.0
EDGE\tPWRN-a-1-1\tPWRN-a-1-2\t1.0
"""
